// Tideland Go Library - Sort
//
// Copyright (C) 2009-2017 Frank Mueller / Tideland / Oldenburg / Germany
//
// All rights reserved. Use of this source code is governed
// by the new BSD license.

// Package sort of the Tideland Go Library provides a parallel quicksort. It
// uses the same interface as the standard sort package.
package sort

// EOF
