// Tideland Go Library - Collections
//
// Copyright (C) 2015-2017 Frank Mueller / Tideland / Oldenburg / Germany
//
// All rights reserved. Use of this source code is governed
// by the new BSD license.

// Package collections of the Tideland Go Library provides some typical and
// often used collection types like a ring buffer, stacks, sets and trees.
// They are implemented as generic collections managing empty interfaces
// as well as typed ones, e.g. for strings. They are not synchronized, so
// this has to be done by the user.
package collections

// EOF
