// Tideland Go Library - Identifier
//
// Copyright (C) 2009-2017 Frank Mueller / Tideland / Oldenburg / Germany
//
// All rights reserved. Use of this source code is governed
// by the new BSD license.

// Package identifier of the Tideland Go Library provides different ways
// to produce identifiers out of diffent input data as well as UUIDs.
//
// The UUID generation can be done according the versions 1, 3, 4, and 5.
// Other identifier types are based on passed data or types. Here
// the individual parts are harmonized and concatenated by the
// passed seperators. It is the users responsibility to check if
// the identifier is unique in its context.
package identifier

// EOF
