// Tideland Go Library - Time Extensions
//
// Copyright (C) 2009-2017 Frank Mueller / Tideland / Oldenburg / Germany
//
// All rights reserved. Use of this source code is governed
// by the new BSD license.

// Package timex helps when working with times and dates. Beside
// tests it contains a crontab for chronological jobs and a retry
// function to let code blocks be retried under well defined conditions
// regarding time and count.
package timex

// EOF
